#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010       Benny Malengier
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Surname class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .secondaryobj import SecondaryObject
from .nameorigintype import NameOriginType
from .const import IDENTICAL, EQUAL, DIFFERENT
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Personal Name
#
# -------------------------------------------------------------------------
class Surname(SecondaryObject):
    """
    Provide surname information of a name.

    A person may have more that one surname in his name
    """

    def __init__(self, source=None, data=None):
        """
        Create a new Surname instance, copying from the source if provided.
        By default a surname is created as primary, use set_primary to change
        """
        if source:
            self.surname = source.surname
            self.prefix = source.prefix
            self.primary = source.primary
            self.origintype = NameOriginType(source.origintype)
            self.connector = source.connector
        else:
            self.surname = ""
            self.prefix = ""
            self.primary = True
            self.origintype = NameOriginType()
            self.connector = ""
        if data:
            self.unserialize(data)

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            self.surname,
            self.prefix,
            self.primary,
            self.origintype.serialize(),
            self.connector,
        )

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Surname"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "surname": {"type": "string", "title": _("Surname")},
                "prefix": {"type": "string", "title": _("Prefix")},
                "primary": {"type": "boolean", "title": _("Primary")},
                "origintype": NameOriginType.get_schema(),
                "connector": {"type": "string", "title": _("Connector")},
            },
        }

    def is_empty(self):
        """
        Indicate if the surname is empty.
        """
        return self.surname == "" and self.prefix == "" and self.connector == ""

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (self.surname, self.prefix, self.primary, origin_type, self.connector) = data
        self.origintype = NameOriginType(origin_type)
        return self

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.surname, self.prefix, self.connector, str(self.origintype)]

    def is_equivalent(self, other):
        """
        Return if this surname is equivalent, that is agrees in type, surname,
        ..., to other.

        :param other: The surname to compare this name to.
        :type other: Surname
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        # TODO what to do with sort and display?
        if (
            self.get_text_data_list() != other.get_text_data_list()
            or self.primary != other.primary
        ):
            return DIFFERENT
        else:
            if self.is_equal(other):
                return IDENTICAL
            else:
                return EQUAL

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this surname.

        Lost: primary, surname, prefix, connector, origintype

        :param acquisition: The surname to merge with the present surname.
        :type acquisition: Surname
        """
        pass

    def get_surname(self):
        """
        Return the surname.

        The surname is one of the not given names coming from the parents
        """
        return self.surname

    def set_surname(self, val):
        """
        Set the surname.

        The surname is one of the not given names coming from the parents
        """
        self.surname = val

    def get_prefix(self):
        """
        Return the prefix (or article) of the surname.

        The prefix is not used for sorting or grouping.
        """
        return self.prefix

    def set_prefix(self, val):
        """
        Set the prefix (or article) of the surname.

        Examples of articles would be 'de' or 'van'.
        """
        self.prefix = val

    def set_origintype(self, the_type):
        """Set the origin type of the Surname instance."""
        self.origintype.set(the_type)

    def get_origintype(self):
        """Return the origin type of the Surname instance."""
        return self.origintype

    def set_connector(self, connector):
        """
        Set the connector for the Surname instance. This defines how a
        surname connects to the next surname (eg in Spanish names).
        """
        self.connector = connector

    def get_connector(self):
        """
        Get the connector for the Surname instance. This defines how a
        surname connects to the next surname (eg in Spanish names).
        """
        return self.connector

    def get_primary(self):
        """Return if this surname is the primary surname"""
        return self.primary

    def set_primary(self, primary=True):
        """
        Set if this surname is the primary surname.replace
        Use :class:`~.surnamebase.SurnameBase` to set the primary surname
        via :meth:`~.surnamebase.SurnameBase.set_primary_surname`

        :param primary: primay surname or not
        :type primary: bool
        """
        self.primary = primary



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2017       Alois Poettker <alois.poettker@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Change IDs of all elements in the database to conform to the
scheme specified in the database's prefix ids
"""

# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------
import re

from gi.repository import Gtk, Gdk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

from gramps.gen.config import config
from gramps.gen.db import DbTxn
from gramps.gen.updatecallback import UpdateCallback

from gramps.gui.display import display_help
from gramps.gui.glade import Glade
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.plug import tool
from gramps.gui.utils import ProgressMeter
from gramps.gui.widgets import MonitoredCheckbox, MonitoredEntry

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Reorder_Gramps_ID", "manual")

PREFIXES = {
    "person": "i",
    "family": "f",
    "event": "e",
    "place": "p",
    "source": "s",
    "citation": "c",
    "repository": "r",
    "media": "o",
    "note": "n",
}
DB_INDXES = {
    "person": "p",
    "family": "f",
    "event": "e",
    "place": "l",
    "source": "s",
    "citation": "c",
    "repository": "r",
    "media": "o",
    "note": "n",
}
# -------------------------------------------------------------------------
#
# Actual tool
#
# -------------------------------------------------------------------------

# gets the prefix, number, suffix specified in a format string, eg:
# P%04dX returns 'P', '04', 'X'  It has to have the integer format with at
# least 3 digits to pass.
_parseformat = re.compile(r"(^[^\d]*)%(0[3-9])d([^\d]*$)")


class ReorderEntry(object):
    """Class for internal values for every primary object"""

    def __init__(self, object_fmt, quant_id, nextgramps_id, obj):
        self.object = obj
        self.width_fmt = 4
        self.object_fmt, self.object_prefix, self.object_suffix = "", "", ""
        self.quant_id, self.actual_id, self.actgramps_id = 0, 0, "0"
        self.calc_id(object_fmt, quant_id)
        self.stored_fmt = self.object_fmt
        self.stored_prefix = self.object_prefix
        self.stored_suffix = self.object_suffix

        self.number_id = int(
            nextgramps_id[
                len(self.object_prefix) : (len(nextgramps_id) - len(self.object_suffix))
            ]
        )
        self.step_id = 1
        self.active_obj, self.change_obj, self.keep_obj = True, False, False

    def set_active(self, active):
        """sets Change flag"""
        self.active_obj = active

    def get_active(self):
        """gets Change flag"""
        return self.active_obj

    def set_fmt(self, object_fmt):
        """sets primary object format"""
        if object_fmt:
            self.calc_id(object_fmt.strip(), self.quant_id)

    def get_fmt(self):
        """gets primary object format"""
        return self.object_fmt

    def res_fmt(self):
        """restore primary object format"""
        return self.stored_fmt

    def set_change(self, change):
        """sets Change flag"""
        self.change_obj = change

    def get_change(self):
        """gets Change flag"""
        return self.change_obj

    def __ret_gid(self, actual):
        """return Gramps ID in correct format"""
        return "%s%s%s" % (
            self.object_prefix,
            str(actual).zfill(self.width_fmt),
            self.object_suffix,
        )

    def calc_id(self, object_fmt, quant_id):
        """calculates identifier prefix, suffix, format & actual value.
        Requires a valid format or returns the default instead"""
        self.object_fmt, self.quant_id = object_fmt, quant_id

        # Default values, ID counting starts with zero!
        formatmatch = _parseformat.match(object_fmt)
        if formatmatch:
            self.object_prefix = formatmatch.groups()[0]
            self.width_fmt = int(formatmatch.groups()[1])
            self.object_suffix = formatmatch.groups()[2]
        else:  # not a legal format string, use default
            self.object_prefix = PREFIXES[self.object].upper()
            self.width_fmt = 4
            self.object_suffix = ""
            self.object_fmt = PREFIXES[self.object].upper() + "%04d"
        self.actgramps_id = self.__ret_gid(self.actual_id)

    def zero_id(self):
        """provide zero Start ID"""
        return self.__ret_gid(0)

    def set_id(self, actual):
        """sets Start ID"""
        text = "".join([i for i in actual.strip() if i in "0123456789"])
        self.actual_id = int(text) if text else 0

    def get_id(self):
        """gets Start ID"""
        return self.__ret_gid(self.actual_id)

    def next_id(self):
        """provide next Start ID"""
        return self.__ret_gid(self.number_id)

    def succ_id(self):
        """provide next actual Gramps ID"""
        self.actual_id += self.step_id
        self.actgramps_id = self.__ret_gid(self.actual_id)

        return self.actgramps_id

    def last_id(self):
        """provide quantities of Gramps IDs"""
        if self.quant_id > 0:
            return self.__ret_gid(self.quant_id - 1)
        else:
            return self.__ret_gid(0)

    def set_step(self, step):
        """sets ID Step width"""
        text = "".join([i for i in step.strip() if i in "0123456789"])
        self.step_id = int(text) if text else 1

    def get_step(self):
        """gets ID Step width"""
        return str(self.step_id)

    def change_step(self, step_entry):
        """change Glade Step entry"""
        step_id = step_entry.get_text().strip()
        if step_id and step_id != str(self.step_id):
            step_entry.set_text(str(self.step_id))

    def set_keep(self, keep):
        """sets Keep flag"""
        self.keep_obj = keep

    def get_keep(self):
        """gets Keep flag"""
        return self.keep_obj


class ReorderIds(tool.BatchTool, ManagedWindow, UpdateCallback):
    """Class for Reodering Gramps ID Tool"""

    xobjects = (
        ("person", "people"),
        ("family", "families"),
        ("event", "events"),
        ("place", "places"),
        ("source", "sources"),
        ("citation", "citations"),
        ("repository", "repositories"),
        ("media", "media"),
        ("note", "notes"),
    )

    def build_menu_names_(self, widget=None):
        """The menu name"""
        return (_("Main window"), _("Reorder Gramps IDs"))

    def __init__(self, dbstate, user, options_class, name, callback=None):
        self.uistate = user.uistate
        self.db = dbstate.db

        if self.uistate:
            tool.BatchTool.__init__(self, dbstate, user, options_class, name)
            if self.fail:
                return  # user denied to modify Gramps IDs

        ManagedWindow.__init__(self, self.uistate, [], self.__class__)
        if not self.uistate:
            UpdateCallback.__init__(self, user.callback)

        self.object_status = True
        self.change_status = False
        self.start_zero = True
        self.step_cnt, self.step_list = 0, ["1", "2", "5", "10"]
        self.keep_status = True

        self.obj_values = {}  # enable access to all internal values
        self.active_entries, self.format_entries = {}, {}
        self.change_entries = {}
        self.start_entries, self.step_entries = {}, {}
        self.keep_entries = {}

        self.prim_methods, self.obj_methods = {}, {}
        for prim_obj, prim_objs in self.xobjects:
            get_handles = "get_%s_handles" % prim_obj
            get_number_obj = "get_number_of_%s" % prim_objs
            prefix_fmt = "%s_prefix" % prim_obj
            get_from_id = "get_%s_from_gramps_id" % prim_obj
            get_from_handle = "get_%s_from_handle" % prim_obj
            next_from_id = "find_next_%s_gramps_id" % prim_obj
            commit = "commit_%s" % prim_obj

            self.prim_methods[prim_obj] = (
                getattr(self.db, prefix_fmt),
                getattr(self.db, get_number_obj)(),
                getattr(self.db, next_from_id)(),
            )
            self.obj_methods[prim_obj] = (
                getattr(self.db, get_handles),
                getattr(self.db, commit),
                getattr(self.db, get_from_id),
                getattr(self.db, get_from_handle),
                getattr(self.db, next_from_id),
            )

            object_fmt, quant_id, next_id = self.prim_methods[prim_obj]

            obj_value = ReorderEntry(object_fmt, quant_id, next_id, prim_obj)
            self.obj_values[prim_obj] = obj_value

        if self.uistate:
            self._display()
        else:
            self._execute()

    def __on_object_button_clicked(self, widget=None):
        """compute all primary objects and toggle the 'Active' attribute"""
        self.object_status = not self.object_status

        for prim_obj, dummy in self.xobjects:
            obj = self.top.get_object("%s_active" % prim_obj)
            obj.set_active(self.object_status)

    def __on_object_button_toggled(self, widget):
        """compute the primary object and toggle the 'Sensitive' attribute"""
        obj_state = widget.get_active()
        obj_name = Gtk.Buildable.get_name(widget).split("_", 1)[0]

        self.active_entries[obj_name].set_val(obj_state)

        for obj_entry in ["actual", "quant", "format", "change"]:
            obj = self.top.get_object("%s_%s" % (obj_name, obj_entry))
            obj.set_sensitive(obj_state)

        for obj_entry in ["start", "step", "keep"]:
            obj = self.top.get_object("%s_change" % obj_name)
            if obj.get_active():
                obj = self.top.get_object("%s_%s" % (obj_name, obj_entry))
                obj.set_sensitive(obj_state)

    def __on_format_button_clicked(self, widget=None):
        """compute all sensitive primary objects and sets the
        'Format' scheme of identifiers"""
        for prim_obj, dummy in self.xobjects:
            obj_format = self.top.get_object("%s_format" % prim_obj)
            if not obj_format.get_sensitive():
                continue

            obj_fmt = self.obj_values[prim_obj].res_fmt()
            self.format_entries[prim_obj].force_value(obj_fmt)
            if self.start_zero:
                obj_id = self.obj_values[prim_obj].zero_id()
            else:
                obj_id = self.obj_values[prim_obj].last_id()
            self.start_entries[prim_obj].force_value(obj_id)

    def __on_change_button_clicked(self, widget=None):
        """compute all primary objects and toggle the 'Change' attribute"""
        self.change_status = not self.change_status

        for prim_obj, dummy in self.xobjects:
            obj_change = self.top.get_object("%s_change" % prim_obj)
            if not obj_change.get_sensitive():
                continue

            self.change_entries[prim_obj].set_val(self.change_status)
            obj_change.set_active(self.change_status)

    def __on_change_button_toggled(self, widget):
        """compute the primary object and toggle the 'Sensitive' attribute"""
        obj_state = widget.get_active()
        obj_name = Gtk.Buildable.get_name(widget).split("_", 1)[0]

        for obj_entry in ["start", "step", "keep"]:
            obj = self.top.get_object("%s_%s" % (obj_name, obj_entry))
            if obj_entry == "keep":
                if (
                    self.obj_values[obj_name].stored_prefix
                    != self.obj_values[obj_name].object_prefix
                    and self.obj_values[obj_name].stored_suffix
                    != self.obj_values[obj_name].object_suffix
                ):
                    self.keep_entries[obj_name].set_val(False)
                else:
                    obj.set_active(obj_state)
                    self.keep_entries[obj_name].set_val(obj_state)
            obj.set_sensitive(obj_state)

    def __on_start_button_clicked(self, widget=None):
        """compute all sensitive primary objects and sets the
        'Start' values of identifiers"""
        self.start_zero = not self.start_zero

        for prim_obj, dummy in self.xobjects:
            obj = self.top.get_object("%s_start" % prim_obj)
            if not obj.get_sensitive():
                continue

            if self.start_zero:
                obj_id = self.obj_values[prim_obj].zero_id()
            else:
                obj_id = self.obj_values[prim_obj].next_id()
            self.start_entries[prim_obj].force_value(obj_id)

    def __on_step_button_clicked(self, widget=None):
        """compute all sensitive primary objects and sets the
        'Step' width of identifiers"""
        self.step_cnt = self.step_cnt + 1 if self.step_cnt < 3 else 0

        for prim_obj, dummy in self.xobjects:
            obj = self.top.get_object("%s_step" % prim_obj)
            if not obj.get_sensitive():
                continue

            step_val = self.step_list[self.step_cnt]
            self.step_entries[prim_obj].force_value(step_val)

    def __on_keep_button_clicked(self, widget=None):
        """compute the primary object and toggle the 'Active' attribute"""
        self.keep_status = not self.keep_status

        for prim_obj, dummy in self.xobjects:
            obj = self.top.get_object("%s_change" % prim_obj)
            if not obj.get_active():
                continue

            obj = self.top.get_object("%s_keep" % prim_obj)
            obj.set_active(self.keep_status)
            self.keep_entries[prim_obj].set_val(self.keep_status)

    def __on_format_entry_keyrelease(self, widget, event, data=None):
        """activated on all return's of an entry"""
        if event.keyval in [Gdk.KEY_Return]:
            obj_name = Gtk.Buildable.get_name(widget).split("_", 1)[0]
            obj_fmt = self.format_entries[obj_name].get_val()
            self.format_entries[obj_name].force_value(obj_fmt)
            self.start_entries[obj_name].update()

            obj_change = self.top.get_object("%s_change" % obj_name)
            obj_change.grab_focus()

        return False

    def __on_format_entry_focusout(self, widget, event, data=None):
        """activated on all focus out of an entry"""
        obj_name = Gtk.Buildable.get_name(widget).split("_", 1)[0]
        obj_fmt = self.format_entries[obj_name].get_val()

        self.format_entries[obj_name].set_text(obj_fmt)
        self.start_entries[obj_name].update()

        return False

    def __on_start_entry_focusout(self, widget, event, data=None):
        """activated on all focus out of an entry"""
        obj_name = Gtk.Buildable.get_name(widget).split("_", 1)[0]
        self.start_entries[obj_name].update()

        return False

    def __on_ok_button_clicked(self, widget=None):
        """execute the reodering and close"""
        self._execute()
        self._update()

        self.close()

    def __on_cancel_button_clicked(self, widget=None):
        """cancel the reodering and close"""
        self.close()

    def __on_help_button_clicked(self, widget=None):
        """display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def _display(self):
        """organize Glade 'Reorder IDs' window"""

        # get the main window from glade
        self.top = Glade(toplevel="reorder-ids")
        window = self.top.toplevel

        # set gramps style title for the window
        self.set_window(window, self.top.get_object("title"), _("Reorder Gramps IDs"))

        # connect signals
        self.top.connect_signals(
            {
                "on_object_button_clicked": self.__on_object_button_clicked,
                "on_object_button_toggled": self.__on_object_button_toggled,
                "on_format_button_clicked": self.__on_format_button_clicked,
                "on_start_button_clicked": self.__on_start_button_clicked,
                "on_step_button_clicked": self.__on_step_button_clicked,
                "on_keep_button_clicked": self.__on_keep_button_clicked,
                "on_change_button_clicked": self.__on_change_button_clicked,
                "on_change_button_toggled": self.__on_change_button_toggled,
                "on_format_entry_keyrelease": self.__on_format_entry_keyrelease,
                "on_format_entry_focusout": self.__on_format_entry_focusout,
                "on_start_entry_focusout": self.__on_start_entry_focusout,
                "on_help_button_clicked": self.__on_help_button_clicked,
                "on_cancel_button_clicked": self.__on_cancel_button_clicked,
                "on_ok_button_clicked": self.__on_ok_button_clicked,
            }
        )

        # Calculate all entries and update Glade window
        for prim_obj, dummy in self.xobjects:
            # populate Object, Actual & Quantity fields with values
            obj_active = self.top.get_object("%s_active" % prim_obj)
            self.active_entries[prim_obj] = MonitoredCheckbox(
                obj_active,
                obj_active,
                self.obj_values[prim_obj].set_active,
                self.obj_values[prim_obj].get_active,
            )
            obj_actual = self.top.get_object("%s_actual" % prim_obj)
            obj_actual.set_text("%s" % self.obj_values[prim_obj].last_id())
            obj_quant = self.top.get_object("%s_quant" % prim_obj)
            obj_quant.set_text("%s" % str(self.obj_values[prim_obj].quant_id))

            # connect/populate Format, Start, Step, Keep & Change fields
            #  with GTK/values
            obj_format = self.top.get_object("%s_format" % prim_obj)
            self.format_entries[prim_obj] = MonitoredEntry(
                obj_format,
                self.obj_values[prim_obj].set_fmt,
                self.obj_values[prim_obj].get_fmt,
            )
            obj_change = self.top.get_object("%s_change" % prim_obj)
            self.change_entries[prim_obj] = MonitoredCheckbox(
                obj_change,
                obj_change,
                self.obj_values[prim_obj].set_change,
                self.obj_values[prim_obj].get_change,
            )
            obj_start = self.top.get_object("%s_start" % prim_obj)
            self.start_entries[prim_obj] = MonitoredEntry(
                obj_start,
                self.obj_values[prim_obj].set_id,
                self.obj_values[prim_obj].get_id,
            )
            obj_step = self.top.get_object("%s_step" % prim_obj)
            self.step_entries[prim_obj] = MonitoredEntry(
                obj_step,
                self.obj_values[prim_obj].set_step,
                self.obj_values[prim_obj].get_step,
                changed=self.obj_values[prim_obj].change_step,
            )
            obj_keep = self.top.get_object("%s_keep" % prim_obj)
            self.keep_entries[prim_obj] = MonitoredCheckbox(
                obj_keep,
                obj_keep,
                self.obj_values[prim_obj].set_keep,
                self.obj_values[prim_obj].get_keep,
                readonly=True,
            )

        # fetch the popup menu
        self.menu = self.top.get_object("popup_menu")

        # ok, let's see what we've done
        self.window.resize(700, 410)
        self.show()

    def _update(self):
        """store changed objects formats in DB"""

        update = False
        for prim_obj, dummy in self.xobjects:
            obj_value = self.obj_values[prim_obj]
            if obj_value.object_fmt != obj_value.stored_fmt:
                constant = "preferences.%sprefix" % PREFIXES[prim_obj]
                config.set(constant, obj_value.object_fmt)
                update = True

        if update:
            config.save()
            self.db.set_prefixes(
                config.get("preferences.iprefix"),
                config.get("preferences.oprefix"),
                config.get("preferences.fprefix"),
                config.get("preferences.sprefix"),
                config.get("preferences.cprefix"),
                config.get("preferences.pprefix"),
                config.get("preferences.eprefix"),
                config.get("preferences.rprefix"),
                config.get("preferences.nprefix"),
            )

    def _execute(self):
        """execute all primary objects and reorder if neccessary"""

        # Update progress calculation
        if self.uistate:
            self.progress = ProgressMeter(_("Reorder Gramps IDs"), "")
        else:
            total_objs = 0
            for prim_obj, dummy in self.xobjects:
                if self.obj_values[prim_obj].active_obj:
                    total_objs += self.obj_values[prim_obj].quant_id
            self.set_total(total_objs)

        # Update database
        self.db.disable_signals()
        for prim_obj, prim_objs in self.xobjects:
            with DbTxn(
                _("Reorder %s IDs ...") % prim_obj, self.db, batch=True
            ) as self.trans:
                if self.obj_values[prim_obj].active_obj:
                    if self.uistate:
                        self.progress.set_pass(
                            _("Reorder %s IDs ...") % _(prim_objs.title()),
                            self.obj_values[prim_obj].quant_id,
                        )
                    # reset the db next_id index to zero so we restart new IDs
                    # at lowest possible position
                    setattr(self.db, DB_INDXES[prim_obj] + "map_index", 0)
                    # Process reordering
                    self._reorder(prim_obj)

        self.db.enable_signals()
        self.db.request_rebuild()

        # Update progress calculation
        if self.uistate:
            self.progress.close()
        else:
            print("\nDone.")

    # finds integer portion in a GrampsID
    _findint = re.compile(r"^[^\d]*(\d+)[^\d]*$")
    # finds prefix, number, suffix of a Gramps ID ignoring a leading or
    # trailing space.  The number must be at least three digits.
    _prob_id = re.compile(r"^ *([^\d]*)(\d{3,9})([^\d]*) *$")

    def _reorder(self, prim_obj):
        """reorders all selected objects with a (new) style, start & step"""

        dup_ids = []  # list of duplicate identifiers
        new_ids = {}  # list of new identifiers

        (
            get_handles,
            commit,
            get_from_id,
            get_from_handle,
            next_from_id,
        ) = self.obj_methods[prim_obj]

        prefix_fmt = self.obj_values[prim_obj].get_fmt()
        prefix = self.obj_values[prim_obj].object_prefix
        suffix = self.obj_values[prim_obj].object_suffix
        old_pref = self.obj_values[prim_obj].stored_prefix
        old_suff = self.obj_values[prim_obj].stored_suffix
        new_id = self.obj_values[prim_obj].get_id()
        keep_fmt = self.obj_values[prim_obj].get_keep()
        change = self.obj_values[prim_obj].get_change()
        index_max = int("9" * self.obj_values[prim_obj].width_fmt)
        do_same = False
        # Process in handle order, which is in order handles were created.
        # This makes renumberd IDs more consistant.
        handles = get_handles()
        handles.sort()

        for handle in handles:
            # Update progress
            if self.uistate:
                self.progress.step()
            else:
                self.update()

            # extract basic data out of the database
            obj = get_from_handle(handle)

            act_id = obj.get_gramps_id()
            # here we see if the ID looks like a new or previous or default
            # Gramps ID.
            # If not we ask user if he really wants to replace it.
            # This should allow user to protect a GetGov ID or similar
            match = self._prob_id.match(act_id)
            if (
                not (
                    match
                    and (
                        prefix == match.groups()[0]
                        and suffix == match.groups()[2]
                        or old_pref == match.groups()[0]
                        and old_suff == match.groups()[2]
                        or len(match.groups()[0]) == 1
                        and len(match.groups()[2]) == 0
                    )
                )
                and not do_same
            ):
                xml = Glade(toplevel="dialog")

                top = xml.toplevel
                # self.top.set_icon(ICON)
                top.set_title("%s - Gramps" % _("Reorder Gramps IDs"))
                apply_to_rest = xml.get_object("apply_to_rest")

                label1 = xml.get_object("toplabel")
                label1.set_text(
                    '<span weight="bold" size="larger">%s</span>'
                    % _("Reorder Gramps IDs")
                )
                label1.set_use_markup(True)

                label2 = xml.get_object("mainlabel")
                label2.set_text(_("Do you want to replace %s?" % act_id))
                top.set_transient_for(self.progress._ProgressMeter__dialog)
                self.progress._ProgressMeter__dialog.set_modal(False)
                top.show()
                response = top.run()
                do_same = apply_to_rest.get_active()
                top.destroy()
                self.progress._ProgressMeter__dialog.set_modal(True)
                if response != Gtk.ResponseType.YES:
                    continue

            elif not match and do_same and response != Gtk.ResponseType.YES:
                continue

            if change:
                # update the defined ID numbers into objects under
                # consideration of keeping ID if format not matches prefix
                # (implication logical boolean operator below)
                if (
                    act_id.startswith(prefix)
                    and act_id.endswith(suffix)
                    or not keep_fmt
                ):
                    obj.set_gramps_id(new_id)
                    commit(obj, self.trans)
                    new_id = self.obj_values[prim_obj].succ_id()
            else:
                # attempt to extract integer - if we can't, treat it as a
                # duplicate
                try:
                    match = self._findint.match(act_id)
                    if match:
                        # get the integer, build the new handle. Make sure it
                        # hasn't already been chosen. If it has, put this
                        # in the duplicate handle list

                        index = int(match.groups()[0])
                        if index > index_max:
                            new_id = next_from_id()
                        else:
                            new_id = prefix_fmt % index

                        if new_id == act_id:
                            if new_id in new_ids:
                                dup_ids.append(obj.get_handle())
                            else:
                                new_ids[new_id] = act_id
                        elif get_from_id(new_id) is not None:
                            dup_ids.append(obj.get_handle())
                        else:
                            obj.set_gramps_id(new_id)
                            commit(obj, self.trans)
                            new_ids[new_id] = act_id
                    else:
                        dup_ids.append(handle)
                except:
                    dup_ids.append(handle)

        # go through the duplicates, looking for the first available
        # handle that matches the new scheme.
        if dup_ids:
            if self.uistate:
                self.progress.set_pass(
                    _("Finding and assigning unused IDs."), len(dup_ids)
                )
            for handle in dup_ids:
                obj = get_from_handle(handle)
                obj.set_gramps_id(next_from_id())
                commit(obj, self.trans)


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class ReorderIdsOptions(tool.ToolOptions):
    """Defines options and provides handling interface."""

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Location class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .secondaryobj import SecondaryObject
from .locationbase import LocationBase
from .const import IDENTICAL, DIFFERENT
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Location class for Places
#
# -------------------------------------------------------------------------
class Location(SecondaryObject, LocationBase):
    """
    Provide information about a place.

    The data including street, locality, city, county, state, and country.
    Multiple Location objects can represent the same place, since names
    of cities, counties, states, and even countries can change with time.
    """

    def __init__(self, source=None):
        """
        Create a Location object, copying from the source object if it exists.
        """
        LocationBase.__init__(self, source)
        if source:
            self.parish = source.parish
        else:
            self.parish = ""

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (LocationBase.serialize(self), self.parish)

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (lbase, self.parish) = data
        LocationBase.unserialize(self, lbase)
        return self

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Location"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "street": {"type": "string", "title": _("Street")},
                "locality": {"type": "string", "title": _("Locality")},
                "city": {"type": "string", "title": _("City")},
                "county": {"type": "string", "title": _("County")},
                "state": {"type": "string", "title": _("State")},
                "country": {"type": "string", "title": _("Country")},
                "postal": {"type": "string", "title": _("Postal Code")},
                "phone": {"type": "string", "title": _("Phone")},
                "parish": {"type": "string", "title": _("Parish")},
            },
        }

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.parish] + LocationBase.get_text_data_list(self)

    def is_equivalent(self, other):
        """
        Return if this location is equivalent to other.

        :param other: The location to compare this one to.
        :type other: Location
        :returns: Constant inidicating degree of equivalence.
        :rtype: int
        """
        if self.is_equal(other):
            return IDENTICAL
        else:
            return DIFFERENT

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this location.

        Lost: everything of acquisition.

        :param acquisition: The location to merge with the present location.
        :type acquisition: Location
        """
        pass

    def is_empty(self):
        return (
            not self.street
            and not self.locality
            and not self.city
            and not self.county
            and not self.state
            and not self.country
            and not self.postal
            and not self.phone
        )

    def set_parish(self, data):
        """Set the religious parish name."""
        self.parish = data

    def get_parish(self):
        """Get the religious parish name."""
        return self.parish



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011       Robert Cheramy
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# RelatedWith
#
# -------------------------------------------------------------------------
class IsRelatedWith(Rule):
    """Rule that checks if a person is related to a specified person"""

    labels = [_("ID:")]
    name = _("People related to <Person>")
    category = _("Relationship filters")
    description = _("Matches people related to a specified person")

    def prepare(self, db, user):
        """prepare so the rule can be executed efficiently
        we build the list of people related to <person> here,
        so that apply is only a check into this list
        """
        self.db = db

        self.relatives = []
        self.add_relative(db.get_person_from_gramps_id(self.list[0]))

    def reset(self):
        self.relatives = []

    def apply(self, db, person):
        return person.handle in self.relatives

    def add_relative(self, start):
        """Non-recursive function that scans relatives and add them to self.relatives"""
        if not (start):
            return

        expand = [start]
        relatives = {}

        while expand:
            person = expand.pop()
            # Add the relative to the list
            if person is None or (person.handle in relatives):
                continue
            relatives[person.handle] = True

            for family_handle in person.get_parent_family_handle_list():
                family = self.db.get_family_from_handle(family_handle)
                if family:
                    # Check Parents
                    for parent_handle in (
                        family.get_father_handle(),
                        family.get_mother_handle(),
                    ):
                        if parent_handle:
                            expand.append(self.db.get_person_from_handle(parent_handle))
                    # Check Sibilings
                    for child_ref in family.get_child_ref_list():
                        expand.append(self.db.get_person_from_handle(child_ref.ref))

            for family_handle in person.get_family_handle_list():
                family = self.db.get_family_from_handle(family_handle)
                if family:
                    # Check Spouse
                    for parent_handle in (
                        family.get_father_handle(),
                        family.get_mother_handle(),
                    ):
                        if parent_handle:
                            expand.append(self.db.get_person_from_handle(parent_handle))
                    # Check Children
                    for child_ref in family.get_child_ref_list():
                        expand.append(self.db.get_person_from_handle(child_ref.ref))

        self.relatives = list(relatives.keys())
        return



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import MatchesFilterBase


# -------------------------------------------------------------------------
#
# MatchesFilter
#
# -------------------------------------------------------------------------
class MatchesPersonFilter(MatchesFilterBase):
    """
    Rule that checks against another filter.

    This is a base rule for subclassing by specific objects.
    Subclasses need to define the namespace class attribute.

    """

    labels = [_("Person filter name:"), _("Include Family events:")]
    name = _("Events of persons matching the <person filter>")
    description = _(
        "Matches events of persons matched by the specified " "person filter name"
    )
    category = _("General filters")

    # we want to have this filter show person filters
    namespace = "Person"

    def prepare(self, db, user):
        MatchesFilterBase.prepare(self, db, user)

        try:
            if int(self.list[1]):
                self.MPF_famevents = True
            else:
                self.MPF_famevents = False
        except IndexError:
            self.MPF_famevents = False

    def apply(self, db, event):
        filt = self.find_filter()
        if filt:
            for classname, handle in db.find_backlink_handles(
                event.get_handle(), ["Person"]
            ):
                if filt.check(db, handle):
                    return True
            if self.MPF_famevents:
                # also include if family event of the person
                for classname, handle in db.find_backlink_handles(
                    event.get_handle(), ["Family"]
                ):
                    family = db.get_family_from_handle(handle)
                    if family.father_handle and filt.check(db, family.father_handle):
                        return True
                    if family.mother_handle and filt.check(db, family.mother_handle):
                        return True

        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = ["LinkBox"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.linkbox")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk


# -------------------------------------------------------------------------
#
# LinkBox class
#
# -------------------------------------------------------------------------
class LinkBox(Gtk.Box):
    def __init__(self, link, button):
        Gtk.Box.__init__(self)
        self.set_spacing(6)
        self.pack_start(link, False, True, 0)
        if button:
            self.pack_start(button, False, True, 0)
        self.show()



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009    Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Place List View
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.plugins.lib.libplaceview import PlaceBaseView
from gramps.gui.views.treemodels.placemodel import PlaceListModel

# -------------------------------------------------------------------------
#
# Internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# PlaceListView
#
# -------------------------------------------------------------------------
class PlaceListView(PlaceBaseView):
    """
    Flat place view.  (Original code in PlaceBaseView).
    """

    def __init__(self, pdata, dbstate, uistate):
        PlaceBaseView.__init__(
            self, pdata, dbstate, uistate, _("Place View"), PlaceListModel, nav_group=0
        )

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide merge capabilities for notes.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.const import URL_MANUAL_SECT3
from ..display import display_help
from ..managedwindow import ManagedWindow
from ..widgets.styledtextbuffer import StyledTextBuffer
from gramps.gen.merge import MergeNoteQuery

# -------------------------------------------------------------------------
#
# Gramps constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = URL_MANUAL_SECT3
WIKI_HELP_SEC = _("Merge_Notes", "manual")
_GLADE_FILE = "mergenote.glade"


# -------------------------------------------------------------------------
#
# MergeNote
#
# -------------------------------------------------------------------------
class MergeNote(ManagedWindow):
    """
    Displays a dialog box that allows two notes to be combined into one.
    """

    def __init__(self, dbstate, uistate, track, handle1, handle2):
        ManagedWindow.__init__(self, uistate, track, self.__class__)
        self.dbstate = dbstate
        database = dbstate.db
        self.no1 = database.get_note_from_handle(handle1)
        self.no2 = database.get_note_from_handle(handle2)

        self.define_glade("mergenote", _GLADE_FILE)
        self.set_window(
            self._gladeobj.toplevel, self.get_widget("note_title"), _("Merge Notes")
        )
        self.setup_configs("interface.merge-note", 600, 250)

        # Detailed selection widgets
        text1 = self.no1.get_styledtext()
        tv1 = self.get_widget("text1")
        tb1 = StyledTextBuffer()
        tv1.set_buffer(tb1)
        tb1.set_text(text1)
        tv1.set_size_request(-1, 60)
        text2 = self.no2.get_styledtext()
        tv2 = self.get_widget("text2")
        tb2 = StyledTextBuffer()
        tv2.set_buffer(tb2)
        tb2.set_text(text2)
        tv2.set_size_request(-1, 60)
        if text1 == text2:
            for widget_name in ("text1", "text2", "text_btn1", "text_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        entry1 = self.get_widget("type1")
        entry2 = self.get_widget("type2")
        entry1.set_text(str(self.no1.get_type()))
        entry2.set_text(str(self.no2.get_type()))
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("type1", "type2", "type_btn1", "type_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        format_names = (_("flowed"), _("preformatted"))
        entry1 = self.get_widget("format1")
        entry2 = self.get_widget("format2")
        entry1.set_text(format_names[self.no1.get_format()])
        entry2.set_text(format_names[self.no2.get_format()])
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("format1", "format2", "format_btn1", "format_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        gramps1 = self.no1.get_gramps_id()
        gramps2 = self.no2.get_gramps_id()
        entry1 = self.get_widget("gramps1")
        entry2 = self.get_widget("gramps2")
        entry1.set_text(gramps1)
        entry2.set_text(gramps2)
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("gramps1", "gramps2", "gramps_btn1", "gramps_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        # Main window widgets that determine which handle survives
        rbutton1 = self.get_widget("handle_btn1")
        rbutton_label1 = self.get_widget("label_handle_btn1")
        rbutton_label2 = self.get_widget("label_handle_btn2")
        text1short = self.no1.get()
        if len(text1short) > 50:
            text1short = text1short[0:47] + "..."
        text2short = self.no2.get()
        if len(text2short) > 50:
            text2short = text2short[0:47] + "..."
        rbutton_label1.set_label("%s [%s]" % (text1short, gramps1))
        rbutton_label2.set_label("%s [%s]" % (text2short, gramps2))
        rbutton1.connect("toggled", self.on_handle1_toggled)

        self.connect_button("note_help", self.cb_help)
        self.connect_button("note_ok", self.cb_merge)
        self.connect_button("note_cancel", self.close)
        self.show()

    def on_handle1_toggled(self, obj):
        """preferred note changes"""
        if obj.get_active():
            self.get_widget("text_btn1").set_active(True)
            self.get_widget("type_btn1").set_active(True)
            self.get_widget("format_btn1").set_active(True)
            self.get_widget("gramps_btn1").set_active(True)
        else:
            self.get_widget("text_btn2").set_active(True)
            self.get_widget("type_btn2").set_active(True)
            self.get_widget("format_btn2").set_active(True)
            self.get_widget("gramps_btn2").set_active(True)

    def cb_help(self, obj):
        """Display the relevant portion of the Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def cb_merge(self, obj):
        """
        Perform the merge of the notes when the merge button is clicked.
        """
        use_handle1 = self.get_widget("handle_btn1").get_active()
        if use_handle1:
            phoenix = self.no1
            titanic = self.no2
        else:
            phoenix = self.no2
            titanic = self.no1

        if self.get_widget("text_btn1").get_active() ^ use_handle1:
            phoenix.set_styledtext(titanic.get_styledtext())
        if self.get_widget("type_btn1").get_active() ^ use_handle1:
            phoenix.set_type(titanic.get_type())
        if self.get_widget("format_btn1").get_active() ^ use_handle1:
            phoenix.set_format(titanic.get_format())
        if self.get_widget("gramps_btn1").get_active() ^ use_handle1:
            phoenix.set_gramps_id(titanic.get_gramps_id())

        query = MergeNoteQuery(self.dbstate, phoenix, titanic)
        query.execute()
        # Add the selected handle to history so that when merge is complete,
        # phoenix is the selected row.
        self.uistate.set_active(phoenix.get_handle(), "Note")
        self.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013       John Ralls <jralls@ceridwen.us>
# Copyright (C) 2020       Nick Hall <nick-h@gramps-project.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import sys
import os
import logging

LOG = logging.getLogger("ResourcePath")
_hdlr = logging.StreamHandler()
_hdlr.setFormatter(logging.Formatter(fmt="%(name)s.%(levelname)s: %(message)s"))
LOG.addHandler(_hdlr)

from ..constfunc import get_env_var


class ResourcePath:
    """
    ResourcePath is a singleton, meaning that only one of them is ever
    created.  At startup it finds the paths to Gramps's resource files and
    caches them for future use.

    It should be called only by const.py; other code should retrieve the
    paths from there.

    Attempt to derive the resource path from the package path assuming that
    one of the three main installation schemes has been used.

    The package path will be one of the following:

    <prefix>/lib/pythonX.Y/site-packages
    <prefix>\\Lib\\site-packages

    <home>/lib/python

    <userbase>/lib/pythonX.Y/site-packages
    <userbase>\\PythonXY\\site-packages

    Where <prefix>, <home> and <userbase> are the resource paths used in the
    Prefix, Home and User installation schemes.

    The use of the command line option "--install-data" in the setup script
    is no longer supported.
    """

    instance = None

    def __new__(cls):
        if not cls.instance:
            cls.instance = super(ResourcePath, cls).__new__(cls)
            cls.instance.initialized = False

        return cls.instance

    def __init__(self):
        if self.initialized:
            return
        package_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "..", "..")
        )
        installed = not os.path.exists(os.path.join(package_path, ".git"))
        if installed:
            test_path = os.path.join("gramps", "authors.xml")
        else:
            test_path = os.path.join("data", "authors.xml")
        resource_path = None
        tmp_path = get_env_var("GRAMPS_RESOURCES")
        if tmp_path and os.path.exists(os.path.join(tmp_path, test_path)):
            resource_path = tmp_path
        elif installed:
            base_path = None
            head, tail = os.path.split(package_path)
            if tail in ("site-packages", "dist-packages"):
                # Prefix or User installation scheme
                head, tail = os.path.split(head)
                if tail.startswith("python"):
                    base_path, tail = os.path.split(head)
                elif tail == "Lib" or tail.startswith("Python"):
                    base_path = head
            elif tail == "python":
                # Home installation scheme
                base_path, tail = os.path.split(head)
            if base_path is not None:
                resource_path = os.path.join(base_path, "share")
            else:
                LOG.error("Unable to determine resource path")
                sys.exit(1)
        else:
            # Let's try to run from source without env['GRAMPS_RESOURCES']:
            resource_path = package_path

        if not os.path.exists(os.path.join(resource_path, test_path)):
            LOG.error("Resource Path %s is invalid", resource_path)
            sys.exit(1)

        resource_path = os.path.abspath(resource_path)
        if installed:
            self.locale_dir = os.path.join(resource_path, "locale")
            self.data_dir = os.path.join(resource_path, "gramps")
            self.image_dir = os.path.join(resource_path, "gramps", "images")
            self.doc_dir = os.path.join(resource_path, "doc", "gramps")

        else:
            self.locale_dir = os.path.join(resource_path, "build", "mo")
            self.image_dir = os.path.join(resource_path, "images")
            self.data_dir = os.path.join(resource_path, "data")
            self.doc_dir = os.path.join(resource_path, "build", "data")

        self.initialized = True



#                                                     -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

#
# Written by Lorenzo Cappelletti <lorenzo.cappelletti@email.it>, 2003
#            Benny Malengier <benny.malengier@gramps-project.org, 2007
#            Maria-Cristina Ciocci <see above>, 2007
#
"""
Italian-Specific classes for relationships.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------
#
# Shared constants
#
# -------------------------------------------------------------------------

_level = [
    "",
    "prim%(gen)s",
    "second%(gen)s",
    "terz%(gen)s",
    "quart%(gen)s",
    "quint%(gen)s",
    "sest%(gen)s",
    "settim%(gen)s",
    "ottav%(gen)s",
    "non%(gen)s",
    "decim%(gen)s",
    "undicesim%(gen)s",
    "dodicesim%(gen)s",
    "tredicesim%(gen)s",
    "quattordicesim%(gen)s",
    "quindicesim%(gen)s",
    "sedicesim%(gen)s",
    "diciasettesim%(gen)s",
    "diciottesim%(gen)s",
    "diciannovesim%(gen)s",
    "ventesim%(gen)s",
]

_level_m = [
    "",
    "primo",
    "secondo",
    "terzo",
    "quarto",
    "quinto",
    "sesto",
    "settimo",
    "ottavo",
    "nono",
    "decimo",
    "undicesimo",
    "dodicesimo",
    "tredicesimo",
    "quattordicesimo",
    "quindicesimo",
    "sedicesimo",
    "diciasettesimo",
    "diciottesimo",
    "diciannovesimo",
    "ventesimo",
]

_level_f = [
    "",
    "prima",
    "seconda",
    "terza",
    "quarta",
    "quinta",
    "sesta",
    "settima",
    "ottava",
    "nona",
    "decima",
    "undicesima",
    "dodicesima",
    "tredicesima",
    "quattordicesima",
    "quindicesima",
    "sedicesima",
    "diciasettesima",
    "diciottesima",
    "diciannovesima",
    "ventesima",
]

_father_level = [
    "",
    "il padre%(step)s%(inlaw)s",
    "il nonno%(step)s%(inlaw)s",
    "il bisnonno%(step)s%(inlaw)s",
    "il trisnonno%(step)s%(inlaw)s",
]

_mother_level = [
    "",
    "la madre%(step)s%(inlaw)s",
    "la nonna%(step)s%(inlaw)s",
    "la bisnonna%(step)s%(inlaw)s",
    "la trisnonna%(step)s%(inlaw)s",
]

_son_level = [
    "",
    "il figlio%(step)s%(inlaw)s",
    "il nipote%(step)s%(inlaw)s diretto",
    "il pronipote%(step)s%(inlaw)s diretto",
]

_daughter_level = [
    "",
    "la figlia%(step)s%(inlaw)s",
    "la nipote%(step)s%(inlaw)s diretta",
    "la pronipote%(step)s%(inlaw)s diretta",
]

_brother_level = [
    "",
    "il fratello%(step)s%(inlaw)s",
    "lo zio%(step)s%(inlaw)s",
    "il prozio%(step)s%(inlaw)s",
]

_sister_level = [
    "",
    "la sorella%(step)s%(inlaw)s",
    "la zia%(step)s%(inlaw)s",
    "la prozia%(step)s%(inlaw)s",
]

_nephew_level = ["", "il nipote%(step)s%(inlaw)s", "il pronipote%(step)s%(inlaw)s"]

_niece_level = ["", "la nipote%(step)s%(inlaw)s", "la pronipote%(step)s%(inlaw)s"]


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    INLAW = " acquisit%(gen)s"

    STEP = " adottiv%(gen)s"

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    # -------------------------------------------------------------------------
    #
    # Specific relationship functions
    #
    # To be honest, I doubt that this relationship naming method is widely
    # spread... If you know of a rigorous, italian naming convention,
    # please, drop me an email.
    #
    # -------------------------------------------------------------------------

    def __gen_suffix(self, gender):
        if gender == Person.MALE:
            return "o"
        return "a"

    def get_parents(self, level):
        if level > len(_level) - 1:
            return "remote ancestors"
        else:
            return "%si genitori" % _level[level]

    def get_father(self, level, step="", inlaw=""):
        gen = "o"

        if level < len(_father_level):
            return _father_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
        elif level < len(_level):
            return (
                "il nonno%(step)s%(inlaw)s della %(level_f)s generazione"
                % {"level_f": _level_f[level], "step": step, "inlaw": inlaw}
                % {"gen": gen}
            )
        else:
            return (
                "l'avo%(step)s%(inlaw)s (%(level)d generazioni)"
                % {"step": step, "inlaw": inlaw, "level": level}
                % {"gen": gen}
            )

    def get_mother(self, level, step="", inlaw=""):
        gen = "a"

        if level < len(_father_level):
            return _mother_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
        elif level < len(_level):
            return (
                "la nonna%(step)s%(inlaw)s della %(level_f)s generazione"
                % {"level_f": _level_f[level], "step": step, "inlaw": inlaw}
                % {"gen": gen}
            )
        else:
            return (
                "l'ava%(step)s%(inlaw)s (%(level)d generazioni)"
                % {"step": step, "inlaw": inlaw, "level": level}
                % {"gen": gen}
            )

    def get_parent_unknown(self, level, step="", inlaw=""):
        gen = "o/a"

        if level == 1:
            return (
                "uno dei genitori%(step)s%(inlaw)s"
                % {"step": step, "inlaw": inlaw}
                % {"gen": gen}
            )
        elif level < len(_father_level):
            return _mother_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
        elif level < len(_level):
            return (
                "nonno/a%(step)s%(inlaw)s della %(level_f)s generazione"
                % {"level_f": _level_f[level], "step": step, "inlaw": inlaw}
                % {"gen": gen}
            )
        else:
            return (
                "l'ava%(step)s%(inlaw)s (%(level)d generazioni)"
                % {"step": step, "inlaw": inlaw, "level": level}
                % {"gen": gen}
            )

    def get_son(self, level, step="", inlaw=""):
        gen = "o"
        if level < len(_son_level):
            return _son_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
        elif level < len(_level):
            return (
                "il nipote%(step)s%(inlaw)s diretto della %(level_f)s generazione"
                % {"level_f": _level_f[level], "step": step, "inlaw": inlaw}
                % {"gen": gen}
            )
        else:
            return (
                "il discendente%(step)s%(inlaw)s diretto (%(level)d generazioni)"
                % {"step": step, "inlaw": inlaw, "level": level}
                % {"gen": gen}
            )

    def get_daughter(self, level, step="", inlaw=""):
        gen = "a"
        if level < len(_daughter_level):
            return (
                _daughter_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
            )
        elif level < len(_level):
            return (
                "la nipote%(step)s%(inlaw)s diretta della %(level_f)s generazione"
                % {"level_f": _level_f[level], "step": step, "inlaw": inlaw}
                % {"gen": gen}
            )
        else:
            return (
                "la discendente%(step)s%(inlaw)s diretta (%(level)d generazioni)"
                % {"step": step, "inlaw": inlaw, "level": level}
                % {"gen": gen}
            )

    def get_uncle(self, level, step="", inlaw=""):
        gen = "o"
        if level < len(_brother_level):
            return _brother_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
        elif level < len(_level):
            return (
                "lo zio%(step)s%(inlaw)s della %(level_f)s generazione"
                % {"level_f": _level_f[level], "step": step, "inlaw": inlaw}
                % {"gen": gen}
            )
        else:
            return (
                "uno zio%(step)s%(inlaw)s lontano (%(level)d generazioni)"
                % {"step": step, "inlaw": inlaw, "level": level}
                % {"gen": gen}
            )

    def get_aunt(self, level, step="", inlaw=""):
        gen = "a"
        if level < len(_brother_level):
            return _sister_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
        elif level < len(_level):
            return (
                "la zia%(step)s%(inlaw)s della %(level_f)s generazione"
                % {"level_f": _level_f[level], "step": step, "inlaw": inlaw}
                % {"gen": gen}
            )
        else:
            return (
                "una zia%(step)s%(inlaw)s lontana (%(level)d generazioni)"
                % {"step": step, "inlaw": inlaw, "level": level}
                % {"gen": gen}
            )

    def get_nephew(self, level, step="", inlaw=""):
        gen = "o"
        if level < len(_nephew_level):
            return _nephew_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
        elif level < len(_level):
            return "il nipote%(step)s%(inlaw)s " "della %(level_f)s generazione" % {
                "level_f": _level_f[level],
                "step": step,
                "inlaw": inlaw,
            } % {"gen": gen}
        else:
            return "un nipote%(step)s%(inlaw)s lontano (" "%(level)d generazioni)" % {
                "step": step,
                "inlaw": inlaw,
                "level": level,
            } % {"gen": gen}

    def get_niece(self, level, step="", inlaw=""):
        gen = "a"
        if level < len(_nephew_level):
            return _niece_level[level] % {"step": step, "inlaw": inlaw} % {"gen": gen}
        elif level < len(_level):
            return "la nipote%(step)s%(inlaw)s " "della %(level_f)s generazione" % {
                "level_f": _level_f[level],
                "step": step,
                "inlaw": inlaw,
            } % {"gen": gen}
        else:
            return "una nipote%(step)s%(inlaw)s lontana (" "%(level)d generazioni)" % {
                "step": step,
                "inlaw": inlaw,
                "level": level,
            } % {"gen": gen}

    def get_male_cousin(self, levelA, levelB, step="", inlaw=""):
        gen = "o"
        return (
            "il cugino%(step)s%(inlaw)s di %(level)d° grado"
            "(%(levA)d-%(levB)d)"
            % {
                "level": levelA + levelB - 1,
                "step": step,
                "inlaw": inlaw,
                "levA": levelA,
                "levB": levelB,
            }
            % {"gen": gen}
        )

    def get_female_cousin(self, levelA, levelB, step="", inlaw=""):
        gen = "a"
        return (
            "la cugina%(step)s%(inlaw)s di %(level)d° grado"
            "(%(levA)d-%(levB)d)"
            % {
                "level": levelA + levelB - 1,
                "step": step,
                "inlaw": inlaw,
                "levA": levelA,
                "levB": levelB,
            }
            % {"gen": gen}
        )

    # -------------------------------------------------------------------------
    #
    # get_relationship
    #
    # -------------------------------------------------------------------------

    def get_relationship(self, db, orig_person, other_person):
        """
        returns a string representing the relationshp between the two people,
        along with a list of common ancestors (typically father, mother)
        """

        if orig_person is None:
            return ("non definito", [])

        if orig_person.get_handle() == other_person.get_handle():
            return ("", [])

        is_spouse = self.is_spouse(db, orig_person, other_person)
        if is_spouse:
            return (is_spouse, [])

        # get_relationship_distance changed, first data is relation to
        # orig person, apperently secondRel in this function
        (secondRel, firstRel, common) = self.get_relationship_distance_new(
            db, orig_person, other_person
        )

        if isinstance(common, str):
            return (common, [])
        elif common:
            person_handle = common[0]
        else:
            return ("", [])

        firstRel = len(firstRel)
        secondRel = len(secondRel)

        if firstRel == 0:
            if secondRel == 0:
                return ("", common)
            elif other_person.get_gender() == Person.MALE:
                return (self.get_father(secondRel), common)
            else:
                return (self.get_mother(secondRel), common)
        elif secondRel == 0:
            if other_person.get_gender() == Person.MALE:
                return (self.get_son(firstRel), common)
            else:
                return (self.get_daughter(firstRel), common)
        elif firstRel == 1:
            if other_person.get_gender() == Person.MALE:
                return (self.get_uncle(secondRel), common)
            else:
                return (self.get_aunt(secondRel), common)
        elif secondRel == 1:
            if other_person.get_gender() == Person.MALE:
                return (self.get_nephew(firstRel - 1), common)
            else:
                return (self.get_niece(firstRel - 1), common)
        else:
            if other_person.get_gender() == Person.MALE:
                return (self.get_male_cousin(firstRel - 1, secondRel - 1), common)
            else:
                return (self.get_female_cousin(firstRel - 1, secondRel - 1), common)

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        See Comment in Relationship Class (relationship.py)
        """

        if only_birth:
            step = ""
        else:
            step = self.STEP

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        if gender_b == Person.MALE:
            rel_str = "un parente%s%s lontano" % (step, inlaw) % {"gen": "o"}
        elif gender_b == Person.FEMALE:
            rel_str = "una parente%s%s lontana" % (step, inlaw) % {"gen": "a"}
        else:
            rel_str = "uno dei parenti%s%s lontani" % (step, inlaw) % {"gen": "i"}

        if Gb == 0:
            if Ga == 0:
                rel_str = "la stessa persona"
            elif Ga == 1 and inlaw and not step:
                if gender_b == Person.MALE:
                    rel_str = "il suocero"
                elif gender_b == Person.FEMALE:
                    rel_str = "la suocera"
                else:
                    rel_str = "uno dei suoceri"
            elif Ga == 1 and not inlaw and step:
                if gender_b == Person.MALE:
                    rel_str = "il patrigno"
                elif gender_b == Person.FEMALE:
                    rel_str = "la matrigna"
                else:
                    rel_str = "uno dei genitori adottivi"
            elif gender_b == Person.MALE:
                rel_str = self.get_father(Ga, step, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self.get_mother(Ga, step, inlaw)
            else:
                rel_str = self.get_parent_unknown(Ga, step, inlaw)
        elif Ga == 0:
            if Gb == 1 and inlaw and not step:
                if gender_b == Person.MALE:
                    rel_str = "il genero"
                elif gender_b == Person.FEMALE:
                    rel_str = "la nuora"
                else:
                    rel_str = "genero/nuora"
            elif gender_b == Person.MALE:
                rel_str = self.get_son(Gb, step, inlaw)
            else:
                rel_str = self.get_daughter(Gb, step, inlaw)
        elif Gb == 1:
            if Ga == 1 and inlaw and not step:
                if gender_b == Person.MALE:
                    rel_str = "il cognato"
                elif gender_b == Person.FEMALE:
                    rel_str = "la cognata"
                else:
                    rel_str = "il cognato/a"
            if gender_b == Person.MALE:
                rel_str = self.get_uncle(Ga, step, inlaw)
            else:
                rel_str = self.get_aunt(Ga, step, inlaw)
        elif Ga == 1:
            if gender_b == Person.MALE:
                rel_str = self.get_nephew(Gb - 1, step, inlaw)
            else:
                rel_str = self.get_niece(Gb - 1, step, inlaw)
        else:
            if gender_b == Person.MALE:
                rel_str = self.get_male_cousin(Gb - 1, Ga - 1, step, inlaw)
            else:
                rel_str = self.get_female_cousin(Gb - 1, Ga - 1, step, inlaw)
        return rel_str

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        """
        Determine the string giving the relation between two siblings of
        type sib_type.
        Eg: b is the brother of a
        Here 'brother' is the string we need to determine
        This method gives more details about siblings than
        get_single_relationship_string can do.

        .. warning:: DON'T TRANSLATE THIS PROCEDURE IF LOGIC IS EQUAL IN YOUR
                     LANGUAGE, AND SAME METHODS EXIST (get_uncle, get_aunt,
                     get_sibling)
        """
        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        if sib_type == self.NORM_SIB or sib_type == self.UNKNOWN_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = self.get_uncle(1, "", "")
                else:
                    rel_str = self.get_aunt(1, "", "")
            else:
                if gender_b == Person.MALE:
                    rel_str = "il cognato"
                elif gender_b == Person.FEMALE:
                    rel_str = "la cognata"
                else:
                    rel_str = "il cognato/a"
        elif (
            sib_type == self.HALF_SIB_FATHER
            or sib_type == self.HALF_SIB_MOTHER
            or sib_type == self.STEP_SIB
        ):
            # Italian has no difference between half and step sibling!
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "il fratellastro"
                elif gender_b == Person.FEMALE:
                    rel_str = "la sorellastra"
                else:
                    rel_str = "il fratellastro/sorellastra"
            else:
                if gender_b == Person.MALE:
                    rel_str = "il fratellastro acquisito"
                elif gender_b == Person.FEMALE:
                    rel_str = "la sorellastra acquisita"
                else:
                    rel_str = "il fratellastro/sorellastra acquisito/a"

        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    #    python src/plugins/rel/rel_it.py

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ._reportdialog import ReportDialog
from gramps.gen.plug.report import CATEGORY_WEB


# -------------------------------------------------------------------------
#
# WebReportDialog class
#
# -------------------------------------------------------------------------
class WebReportDialog(ReportDialog):
    """
    The WebReportDialog base class.  This is a base class for generating
    dialogs for web page reports.
    """

    def __init__(self, dbstate, uistate, option_class, name, trans_name):
        """Initialize a dialog"""
        self.category = CATEGORY_WEB
        ReportDialog.__init__(self, dbstate, uistate, option_class, name, trans_name)
        self.options.handler.set_format_name("html")

    def setup_init(self):
        pass

    def setup_target_frame(self):
        """Target frame is not used."""
        pass

    def parse_target_frame(self):
        """Target frame is not used."""
        return 1



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2008-2009  Stephane Charette <stephanecharette@gmail.com>
#               2009       Gary Burton
#               2011       Robert Cheramy <robert@cheramy.net>
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
from copy import deepcopy

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GdkPixbuf
from gi.repository import Gdk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..utils import open_file_with_default_application
from gramps.gen.const import THUMBSCALE
from gramps.gen.mime import get_description, get_type
from gramps.gen.utils.thumbnails import get_thumbnail_image, find_mime_type_pixbuf
from gramps.gen.utils.file import media_path_full, find_file, create_checksum
from gramps.gen.lib import NoteType
from gramps.gen.db import DbTxn
from ..glade import Glade
from .displaytabs import (
    CitationEmbedList,
    MediaAttrEmbedList,
    MediaBackRefList,
    NoteTab,
)
from ..widgets import (
    MonitoredSpinButton,
    MonitoredEntry,
    PrivacyButton,
    MonitoredDate,
    MonitoredTagList,
    SelectionWidget,
    Region,
)
from .editreference import RefTab, EditReference
from .addmedia import AddMedia
from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = _("Media_Reference_Editor_dialog", "manual")


# -------------------------------------------------------------------------
#
# EditMediaRef
#
# -------------------------------------------------------------------------
class EditMediaRef(EditReference):
    def __init__(self, state, uistate, track, media, media_ref, update):
        EditReference.__init__(self, state, uistate, track, media, media_ref, update)
        if not self.source.get_handle():
            # show the addmedia dialog immediately, with track of parent.
            AddMedia(
                state, self.uistate, self.track, self.source, self._update_addmedia
            )
        else:
            self.original = deepcopy(self.source.serialize())

    def _local_init(self):
        self.top = Glade()
        self.set_window(
            self.top.toplevel, self.top.get_object("title"), _("Media Reference Editor")
        )
        self.setup_configs("interface.media-ref", 600, 450)

        self.define_warn_box(self.top.get_object("warn_box"))
        self.top.get_object("label427").set_text(_("Y", "Y coordinate"))
        self.top.get_object("label428").set_text(_("Y", "Y coordinate"))

        tblref = self.top.get_object("table50")
        self.notebook_ref = self.top.get_object("notebook_ref")
        self.track_ref_for_deletion("notebook_ref")
        self.expander = self.top.get_object("expander1")
        # recreate start page as GrampsTab
        self.notebook_ref.remove_page(0)
        self.reftab = RefTab(
            self.dbstate, self.uistate, self.track, _("General"), tblref
        )
        self.track_ref_for_deletion("reftab")
        tblref = self.top.get_object("table2")
        self.notebook_shared = self.top.get_object("notebook_shared")
        # recreate start page as GrampsTab
        self.notebook_shared.remove_page(0)
        self.track_ref_for_deletion("notebook_shared")
        self.primtab = RefTab(
            self.dbstate, self.uistate, self.track, _("_General"), tblref
        )
        self.track_ref_for_deletion("primtab")
        self.rect_pixbuf = None

    def setup_filepath(self):
        self.select = self.top.get_object("file_select")
        self.track_ref_for_deletion("select")
        self.file_path = self.top.get_object("path")
        self.track_ref_for_deletion("file_path")

        self.file_path.set_text(self.source.get_path())
        self.select.connect("clicked", self.select_file)

    def determine_mime(self):
        descr = get_description(self.source.get_mime_type())
        if descr:
            self.mimetext.set_text(descr)

        path = self.file_path.get_text()
        path_full = media_path_full(self.db, path)
        if path != self.source.get_path() and path_full != self.source.get_path():
            # redetermine mime
            mime = get_type(find_file(path_full))
            self.source.set_mime_type(mime)
            descr = get_description(mime)
            if descr:
                self.mimetext.set_text(descr)
            else:
                self.mimetext.set_text(_("Unknown"))
        # if mime type not set, is note
        if not self.source.get_mime_type():
            self.mimetext.set_text(_("Note"))

    def draw_preview(self):
        """
        Draw the two preview images. This method can be called on eg change of
        the path.
        """
        mtype = self.source.get_mime_type()
        if mtype:
            fullpath = media_path_full(self.db, self.source.get_path())
            pb = get_thumbnail_image(fullpath, mtype)
            self.pixmap.set_from_pixbuf(pb)
            self.selection.load_image(fullpath)
        else:
            pb = find_mime_type_pixbuf("text/plain")
            self.pixmap.set_from_pixbuf(pb)
            self.selection.load_image("")

    def _setup_fields(self):
        ebox_shared = self.top.get_object("eventbox")
        ebox_shared.connect("button-press-event", self.button_press_event)
        self.pixmap = self.top.get_object("pixmap")
        self.mimetext = self.top.get_object("type")
        self.track_ref_for_deletion("mimetext")

        coord = self.source_ref.get_rectangle()
        # upgrade path: set invalid (from eg old db) to none

        if coord is not None and coord in (
            (None,) * 4,
            (0, 0, 100, 100),
            (coord[0], coord[1]) * 2,
        ):
            coord = None

        if coord is not None:
            self.rectangle = coord
        else:
            self.rectangle = (0, 0, 100, 100)

        self.selection = SelectionWidget()
        self.selection.set_multiple_selection(False)
        self.selection.connect("region-modified", self.region_modified)
        self.selection.connect("region-created", self.region_modified)
        self.expander.connect("activate", self.selection.expander)
        frame = self.top.get_object("frame9")
        frame.add(self.selection)
        self.track_ref_for_deletion("selection")
        self.selection.show()

        self.setup_filepath()
        self.determine_mime()

        corners = ["corner1_x", "corner1_y", "corner2_x", "corner2_y"]

        if coord and isinstance(coord, tuple):
            for index, corner in enumerate(corners):
                self.top.get_object(corner).set_value(coord[index])
        else:
            for corner, value in zip(corners, [0, 0, 100, 100]):
                self.top.get_object(corner).set_value(value)

        if self.dbstate.db.readonly:
            for corner in corners:
                self.top.get_object(corner).set_sensitive(False)

        self.corner1_x_spinbutton = MonitoredSpinButton(
            self.top.get_object("corner1_x"),
            self.set_corner1_x,
            self.get_corner1_x,
            self.db.readonly,
        )
        self.track_ref_for_deletion("corner1_x_spinbutton")

        self.corner1_y_spinbutton = MonitoredSpinButton(
            self.top.get_object("corner1_y"),
            self.set_corner1_y,
            self.get_corner1_y,
            self.db.readonly,
        )
        self.track_ref_for_deletion("corner1_y_spinbutton")

        self.corner2_x_spinbutton = MonitoredSpinButton(
            self.top.get_object("corner2_x"),
            self.set_corner2_x,
            self.get_corner2_x,
            self.db.readonly,
        )
        self.track_ref_for_deletion("corner2_x_spinbutton")

        self.corner2_y_spinbutton = MonitoredSpinButton(
            self.top.get_object("corner2_y"),
            self.set_corner2_y,
            self.get_corner2_y,
            self.db.readonly,
        )
        self.track_ref_for_deletion("corner2_y_spinbutton")

        self.descr_window = MonitoredEntry(
            self.top.get_object("description"),
            self.source.set_description,
            self.source.get_description,
            self.db.readonly,
        )

        self.ref_privacy = PrivacyButton(
            self.top.get_object("private"), self.source_ref, self.db.readonly
        )

        self.gid = MonitoredEntry(
            self.top.get_object("gid"),
            self.source.set_gramps_id,
            self.source.get_gramps_id,
            self.db.readonly,
        )

        self.privacy = PrivacyButton(
            self.top.get_object("privacy"), self.source, self.db.readonly
        )

        self.path_obj = MonitoredEntry(
            self.top.get_object("path"),
            self.source.set_path,
            self.source.get_path,
            self.db.readonly,
        )

        self.date_field = MonitoredDate(
            self.top.get_object("date_entry"),
            self.top.get_object("date_edit"),
            self.source.get_date_object(),
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.tags = MonitoredTagList(
            self.top.get_object("tag_label"),
            self.top.get_object("tag_button"),
            self.source.set_tag_list,
            self.source.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

    def _post_init(self):
        """
        Initialization that must happen after the window is shown.
        """
        self.draw_preview()
        self.update_region()

    def set_corner1_x(self, value):
        """
        Callback for the signal handling of the spinbutton for the first
        corner x coordinate of the subsection.
        Updates the subsection thumbnail using the given value

        @param value: the first corner x coordinate of the subsection in int
        """

        self.rectangle = (value,) + self.rectangle[1:]
        self.update_region()

    def set_corner1_y(self, value):
        """
        Callback for the signal handling of the spinbutton for the first
        corner y coordinate of the subsection.
        Updates the subsection thumbnail using the given value

        @param value: the first corner y coordinate of the subsection in int
        """

        self.rectangle = self.rectangle[:1] + (value,) + self.rectangle[2:]
        self.update_region()

    def set_corner2_x(self, value):
        """
        Callback for the signal handling of the spinbutton for the second
        corner x coordinate of the subsection.
        Updates the subsection thumbnail using the given value

        @param value: the second corner x coordinate of the subsection in int
        """

        self.rectangle = self.rectangle[:2] + (value,) + self.rectangle[3:]
        self.update_region()

    def set_corner2_y(self, value):
        """
        Callback for the signal handling of the spinbutton for the second
        corner y coordinate of the subsection.
        Updates the subsection thumbnail using the given value

        @param value: the second corner y coordinate of the subsection in int
        """

        self.rectangle = self.rectangle[:3] + (value,)
        self.update_region()

    def get_corner1_x(self):
        """
        Callback for the signal handling of the spinbutton for the first corner
        x coordinate of the subsection.

        @returns: the first corner x coordinate of the subsection or 0 if
                  there is no selection
        """
        return self.rectangle[0]

    def get_corner1_y(self):
        """
        Callback for the signal handling of the spinbutton for the first corner
        y coordinate of the subsection.

        @returns: the first corner y coordinate of the subsection or 0 if
                  there is no selection
        """
        return self.rectangle[1]

    def get_corner2_x(self):
        """
        Callback for the signal handling of the spinbutton for the second
        corner x coordinate of the subsection.

        @returns: the second corner x coordinate of the subsection or 100 if
                  there is no selection
        """
        return self.rectangle[2]

    def get_corner2_y(self):
        """
        Callback for the signal handling of the spinbutton for the second
        corner x coordinate of the subsection.

        @returns: the second corner x coordinate of the subsection or 100 if
                  there is no selection
        """
        return self.rectangle[3]

    def update_region(self):
        """
        Updates the thumbnail of the specified subsection.
        """
        if not self.selection.is_image_loaded():
            return
        real = self.selection.proportional_to_real_rect(self.rectangle)
        region = Region(real[0], real[1], real[2], real[3])
        self.selection.set_regions([region])
        self.selection.select(region)  # update the selection box shown
        self.selection.refresh()

    def region_modified(self, widget):
        """
        Update new values when the selection is changed.
        """
        real = self.selection.get_selection()
        coords = self.selection.real_to_proportional_rect(real)
        self.corner1_x_spinbutton.set_value(coords[0])
        self.corner1_y_spinbutton.set_value(coords[1])
        self.corner2_x_spinbutton.set_value(coords[2])
        self.corner2_y_spinbutton.set_value(coords[3])

    def build_menu_names(self, person):
        """
        Provide the information needed by the base class to define the
        window management menu entries.
        """
        if self.source:
            submenu_label = _("Media: %s") % self.source.get_gramps_id()
        else:
            submenu_label = _("New Media")
        return (_("Media Reference Editor"), submenu_label)

    def button_press_event(self, obj, event):
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            photo_path = media_path_full(self.db, self.source.get_path())
            open_file_with_default_application(photo_path, self.uistate)

    def _update_addmedia(self, obj):
        """
        Called when the add media dialog has been called.
        This allows us to update the main form in response to
        any changes: Redraw relevant fields: description, mimetype and path
        """
        for obj in (self.descr_window, self.path_obj):
            obj.update()
        self.determine_mime()
        self.update_checksum()
        self.draw_preview()

    def update_checksum(self):
        self.uistate.set_busy_cursor(True)
        media_path = media_path_full(self.dbstate.db, self.source.get_path())
        self.source.set_checksum(create_checksum(os.path.normpath(media_path)))
        self.uistate.set_busy_cursor(False)

    def select_file(self, val):
        self.determine_mime()
        path = self.file_path.get_text()
        self.source.set_path(path)
        AddMedia(
            self.dbstate, self.uistate, self.track, self.source, self._update_addmedia
        )

    def _connect_signals(self):
        self.define_cancel_button(self.top.get_object("button84"))
        self.define_ok_button(self.top.get_object("button82"), self.save)
        # TODO help button (rename glade button name)
        self.define_help_button(
            self.top.get_object("button104"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("media-rebuild", self.close)
        self._add_db_signal("media-delete", self.check_for_close)

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.
        """
        notebook_ref = self.top.get_object("notebook_ref")
        notebook_src = self.top.get_object("notebook_shared")

        self._add_tab(notebook_src, self.primtab)
        self._add_tab(notebook_ref, self.reftab)

        self.srcref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source_ref.get_citation_list(),
            "mediaref_editor_ref_citations",
        )
        self._add_tab(notebook_ref, self.srcref_list)
        self.track_ref_for_deletion("srcref_list")

        self.attr_list = MediaAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source_ref.get_attribute_list(),
            "mediaref_editor_ref_attributes",
        )
        self._add_tab(notebook_ref, self.attr_list)
        self.track_ref_for_deletion("attr_list")

        self.backref_list = MediaBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.source.handle),
            "mediaref_editor_shared_references",
            self.enable_warnbox,
        )
        self._add_tab(notebook_src, self.backref_list)
        self.track_ref_for_deletion("backref_list")

        self.note_ref_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source_ref.get_note_list(),
            "mediaref_editor_ref_notes",
            notetype=NoteType.MEDIAREF,
        )
        self._add_tab(notebook_ref, self.note_ref_tab)
        self.track_ref_for_deletion("note_ref_tab")

        self.src_srcref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_citation_list(),
            "mediaref_editor_shared_citations",
        )
        self._add_tab(notebook_src, self.src_srcref_list)
        self.track_ref_for_deletion("src_srcref_list")

        self.src_attr_list = MediaAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_attribute_list(),
            "mediaref_editor_shared_attributes",
        )
        self._add_tab(notebook_src, self.src_attr_list)
        self.track_ref_for_deletion("src_attr_list")

        self.src_note_ref_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.source.get_note_list(),
            "mediaref_editor_shared_notes",
            notetype=NoteType.MEDIA,
        )
        self._add_tab(notebook_src, self.src_note_ref_tab)
        self.track_ref_for_deletion("src_note_ref_tab")

        self._setup_notebook_tabs(notebook_src)
        self._setup_notebook_tabs(notebook_ref)

    def save(self, *obj):
        # first save primary object
        if self.source.handle:
            # only commit if it has changed
            if self.source.serialize() != self.original:
                with DbTxn(
                    _("Edit Media Object (%s)") % self.source.get_description(), self.db
                ) as trans:
                    self.db.commit_media(self.source, trans)
        else:
            if self.check_for_duplicate_id("Media"):
                return
            with DbTxn(
                _("Add Media Object (%s)") % self.source.get_description(), self.db
            ) as trans:
                self.db.add_media(self.source, trans)

        # save reference object in memory
        coord = (
            self.top.get_object("corner1_x").get_value_as_int(),
            self.top.get_object("corner1_y").get_value_as_int(),
            self.top.get_object("corner2_x").get_value_as_int(),
            self.top.get_object("corner2_y").get_value_as_int(),
        )

        # do not set unset or invalid coord

        if coord is not None and coord in (
            (None,) * 4,
            (0, 0, 100, 100),
            (coord[0], coord[1]) * 2,
        ):
            coord = None

        self.source_ref.set_rectangle(coord)

        # call callback if given
        if self.update:
            self.update(self.source_ref, self.source)
            self.update = None
        self.close()



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2021-      T G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Class:
    AlphabeticIndex - approximate emulation of ICU Alphabetic Index
"""

# ------------------------------------------------
# python modules
# ------------------------------------------------
from unicodedata import normalize
from collections import defaultdict
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

HAVE_ICU = False
try:
    from icu import Locale, Collator

    HAVE_ICU = True
except ImportError:
    try:
        from PyICU import Locale, Collator

        HAVE_ICU = True
    except ImportError:
        pass

LOG = logging.getLogger(".NarrativeWeb")
COLLATE_LANG = glocale.collation


class U_ENUM_OUT_OF_SYNC_ERROR(Exception):  # pylint: disable=invalid-name
    """
    Exception to match the error in the ICU AlphabetiIndex
    """

    pass


# See : http://www.gramps-project.org/bugs/view.php?id = 4423

# Contraction data taken from CLDR 22.1. Only the default variant is considered.
# The languages included below are, by no means, all the langauges that have
# contractions - just a sample of langauges that have been supported

# At the time of writing (Feb 2013), the following langauges have greater that
# 50% coverage of translation of Gramps: bg Bulgarian, ca Catalan, cs Czech, da
# Danish, de German, el Greek, en_GB, es Spanish, fi Finish, fr French, he
# Hebrew, hr Croation, hu Hungarian, it Italian, ja Japanese, lt Lithuanian, nb
# Noregian Bokmål, nn Norwegian Nynorsk, nl Dutch, pl Polish, pt_BR Portuguese
# (Brazil), pt_P Portugeuse (Portugal), ru Russian, sk Slovak, sl Slovenian, sv
# Swedish, vi Vietnamese, zh_CN Chinese.

# Key is the language (or language and country), Value is a list of
# contractions. Each contraction consists of a tuple. First element of the
# tuple is the list of characters, second element is the string to use as the
# index entry.

# The DUCET contractions (e.g. LATIN CAPIAL LETTER L, MIDDLE DOT) are ignored,
# as are the supresscontractions in some locales.

CONTRACTIONS_DICT = {
    # bg Bulgarian validSubLocales="bg_BG" no contractions
    # ca Catalan validSubLocales="ca_AD ca_ES"
    "ca": [(("l·", "L·"), "L")],
    # Czech, validSubLocales="cs_CZ" Czech_Czech Republic
    "cs": [(("ch", "cH", "Ch", "CH"), "CH")],
    # Danish validSubLocales="da_DK" Danish_Denmark
    "da": [(("aa", "Aa", "AA"), "Å")],
    # de German validSubLocales="de_AT de_BE de_CH de_DE de_LI de_LU" no
    # contractions in standard collation.
    # el Greek validSubLocales="el_CY el_GR" no contractions.
    # es Spanish validSubLocales="es_419 es_AR es_BO es_CL es_CO es_CR es_CU
    # es_DO es_EA es_EC es_ES es_GQ es_GT es_HN es_IC es_MX es_NI es_PA es_PE
    # es_PH es_PR es_PY es_SV es_US es_UY es_VE" no contractions in standard
    # collation.
    # fi Finish validSubLocales="fi_FI" no contractions in default (phonebook)
    # collation.
    # fr French no collation data.
    # he Hebrew validSubLocales="he_IL" no contractions
    # hr Croation validSubLocales="hr_BA hr_HR"
    "hr": [
        (("dž", "Dž", "DŽ"), "DŽ"),
        (("lj", "Lj", "LJ"), "Ǉ"),
        (("Nj", "NJ", "nj"), "Ǌ"),
    ],
    # Hungarian hu_HU for two and three character contractions.
    "hu": [
        (("cs", "Cs", "CS"), "CS"),
        (("dzs", "Dzs", "DZS"), "DZS"),  # order is important
        (("dz", "Dz", "DZ"), "DZ"),
        (("gy", "Gy", "GY"), "GY"),
        (("ly", "Ly", "LY"), "LY"),
        (("ny", "Ny", "NY"), "NY"),
        (("sz", "Sz", "SZ"), "SZ"),
        (("ty", "Ty", "TY"), "TY"),
        (("zs", "Zs", "ZS"), "ZS"),
    ],
    # it Italian no collation data.
    # ja Japanese unable to process the data as it is too complex.
    # lt Lithuanian no contractions.
    # Norwegian Bokmål
    "nb": [(("aa", "Aa", "AA"), "Å")],
    # nn Norwegian Nynorsk validSubLocales="nn_NO"
    "nn": [(("aa", "Aa", "AA"), "Å")],
    # nl Dutch no collation data.
    # pl Polish validSubLocales="pl_PL" no contractions
    # pt Portuguese no collation data.
    # ru Russian validSubLocales="ru_BY ru_KG ru_KZ ru_MD ru_RU ru_UA" no
    # contractions
    # Slovak,  validSubLocales="sk_SK" Slovak_Slovakia
    # having DZ in Slovak as a contraction was rejected in
    # http://unicode.org/cldr/trac/ticket/2968
    "sk": [(("ch", "cH", "Ch", "CH"), "Ch")],
    # sl Slovenian validSubLocales="sl_SI" no contractions
    # sv Swedish validSubLocales="sv_AX sv_FI sv_SE" default collation is
    # "reformed" no contractions.
    # vi Vietnamese validSubLocales="vi_VN" no contractions.
    # zh Chinese validSubLocales="zh_Hans zh_Hans_CN zh_Hans_SG" no contractions
    # in Latin characters the others are too complex.
}

# The comment below from the glibc locale sv_SE in
# localedata/locales/sv_SE :
#
# % The letter w is normally not present in the Swedish alphabet. It
# % exists in some names in Swedish and foreign words, but is accounted
# % for as a variant of 'v'.  Words and names with 'w' are in Swedish
# % ordered alphabetically among the words and names with 'v'. If two
# % words or names are only to be distinguished by 'v' or % 'w', 'v' is
# % placed before 'w'.
#
# See : http://www.gramps-project.org/bugs/view.php?id = 2933
#

# HOWEVER: the characters V and W in Swedish are not considered as a special
# case for several reasons. (1) The default collation for Swedish (called the
# 'reformed' collation type) regards the difference between 'v' and 'w' as a
# primary difference. (2) 'v' and 'w' in the 'standard' (non-default) collation
# type are not a contraction, just a case where the difference is secondary
# rather than primary. (3) There are plenty of other languages where a
# difference that is primary in other languages is secondary, and those are not
# specially handled.


def first_letter(string, rlocale=glocale):
    """
    Receives a string and returns the first letter
    """
    if string is None or len(string) < 1:
        return " "

    norm_unicode = normalize("NFKC", str(string))
    contractions = CONTRACTIONS_DICT.get(rlocale.collation)
    if contractions is None:
        contractions = CONTRACTIONS_DICT.get(rlocale.collation.split("_")[0])

    if contractions is not None:
        for contraction in contractions:
            count = len(contraction[0][0])
            if len(norm_unicode) >= count and norm_unicode[:count] in contraction[0]:
                return contraction[1]

    # no special case
    return norm_unicode[0].upper()


if HAVE_ICU:

    def primary_difference(prev_key, new_key, rlocale=glocale):
        """
        Try to use the PyICU collation.
        If we generate a report for another language, make sure we use the good
        collation sequence
        """
        collate_lang = Locale(rlocale.collation)
        collation = Collator.createInstance(collate_lang)
        collation.setStrength(Collator.PRIMARY)
        return collation.compare(prev_key, new_key) != 0

else:

    def primary_difference(prev_key, new_key, rlocale=glocale):
        """
        The PyICU collation is not available.

        Returns true if there is a primary difference between the two parameters
        See http://www.gramps-project.org/bugs/view.php?id=2933#c9317 if
        letter[i]+'a' < letter[i+1]+'b' and letter[i+1]+'a' < letter[i]+'b' is
        true then the letters should be grouped together

        The test characters here must not be any that are used in contractions.
        """

        return rlocale.sort_key(prev_key + "e") >= rlocale.sort_key(
            new_key + "f"
        ) or rlocale.sort_key(new_key + "e") >= rlocale.sort_key(prev_key + "f")


def get_index_letter(letter, index_list, rlocale=glocale):
    """
    This finds the letter in the index_list that has no primary difference from
    the letter provided. See the discussion in get_first_letters above.
    Continuing the example, if letter is Å and index_list is A, then this would
    return A.

    @param: letter     -- The letter to find in the index_list
    @param: index_list -- The list of all first letters in use
    @param: rlocale    -- The locale to use
    """
    for index in index_list:
        if not primary_difference(letter, index, rlocale):
            return index

    LOG.warning("Initial letter '%s' not found in alphabetic navigation list", letter)
    LOG.debug("filtered sorted index list %s", index_list)
    return letter


# ------------------------------------------------------------
#
# AlphabeticIndex (local non-ICU version)
#
# ------------------------------------------------------------
class AlphabeticIndex:
    """
    Approximately emulate the ICU AlphabeticIndex
    """

    def __init__(self, rlocale):
        self.rlocale = rlocale
        self._record_list = []
        # self.index_list = []

        self._bucket_list = []
        self._dirty = False
        self._bucket = -1
        self._record = -1
        # Externally available properties
        self.bucketLabel = ""  # pylint: disable=invalid-name
        self.recordName = ""  # pylint: disable=invalid-name
        self.recordData = ""  # pylint: disable=invalid-name
        self.bucketRecordCount = 0  # pylint: disable=invalid-name

    def __create_index(self):
        """
        Internal routine to create the Alphabetic Index
        """
        # The first letter (or letters if there is a contraction) are extracted
        # from all the records in the _record_list. There may be duplicates, and
        # there may be letters where there is only a secondary or tertiary
        # difference, not a primary difference.
        index_list = []
        for name, dummy_data in self._record_list:
            ltr = first_letter(name, self.rlocale)
            index_list.append(ltr)
        # The list is sorted in collation order.
        index_list.sort(key=self.rlocale.sort_key)
        # For each group with secondary or tertiary differences, the first in
        # collation sequence is retained. For example, assume the default
        # collation sequence (DUCET) and names Ånström and Apple. These will
        # sort in the order shown. Å and A have a secondary difference. If the
        # first letter from these names was chosen then the index entry would be
        # Å. This is not desirable. Instead, the initial letters are extracted
        # (Å and A). These are sorted, which gives A and Å. Then the first of
        # these is used for the index entry.

        # now remove letters where there is not a primary difference
        first = True
        prev_index = None
        for nkey in index_list[:]:
            # iterate over a slice copy
            if first or primary_difference(prev_index, nkey, self.rlocale):
                first = False
                prev_index = nkey
            else:
                index_list.remove(nkey)

        # finally construct the buckets and contents
        bucket_dict = defaultdict(list)
        for name, data in sorted(
            self._record_list, key=lambda x: self.rlocale.sort_key(x[0])
        ):
            letter = first_letter(name, self.rlocale)
            letter = get_index_letter(letter, index_list, self.rlocale)
            bucket_dict[letter].append((name, data))

        self._bucket_list = sorted(
            bucket_dict.items(), key=lambda x: self.rlocale.sort_key(x[0])
        )

        self._dirty = False

    def addRecord(self, name, data):  # pylint: disable=invalid-name
        """
        Add a record to the index.

        Each record will be associated with an index Bucket based on the
        record's name. The list of records for each bucket will be sorted
        based on the collation ordering of the names in the index's locale.
        Records with duplicate names are permitted; they will be kept in the
        order that they were added.

        @param: name        --  The display name
                                for the Record. The Record will be placed in
                                a bucket based on this name.
        @param: data        --  An optional pointer to user data associated
                                with this item. When iterating the contents
                                of a bucket, both the data pointer the name
                                will be available for each Record.
        """
        self._record_list.append((name, data))
        self._dirty = True

    def resetBucketIterator(self):  # pylint: disable=invalid-name
        """
        Reset the Bucket iteration for this index.

        The next call to nextBucket() will restart the iteration at the
        first label.
        """
        if self._dirty:
            self.__create_index()
        self._bucket = -1
        self.bucketLabel = ""
        self._record = -1
        self.recordName = ""
        self.recordData = ""

    def nextBucket(self):  # pylint: disable=invalid-name
        """
        Advance the iteration over the Buckets of this index.

        Return false if there are no more Buckets.
        """
        if self._dirty:
            raise U_ENUM_OUT_OF_SYNC_ERROR
        self._bucket += 1
        self._record = -1
        if self._bucket < len(self._bucket_list):
            self.bucketLabel = self._bucket_list[self._bucket][0]
            self.bucketRecordCount = len(self._bucket_list[self._bucket][1])
            self.recordName = ""
            self.recordData = None
            return True
        else:
            return False

    def nextRecord(self):  # pylint: disable=invalid-name
        """
        Advance to the next record in the current Bucket.

        When nextBucket() is called, Record iteration is reset to just
        before the first Record in the new Bucket.
        """
        if self._dirty:
            raise U_ENUM_OUT_OF_SYNC_ERROR
        self._record += 1
        if self._record < len(self._bucket_list[self._bucket][1]):
            curr_bucket = self._bucket_list[self._bucket]
            bucket_value = curr_bucket[1]
            curr_record = bucket_value[self._record]
            self.recordName = curr_record[0]
            self.recordData = curr_record[1]
            return True
        else:
            self.recordName = ""
            self.recordData = None
            return False
